#include "array.h"
#include <stdlib.h>
#include <string.h>

static int arraylen(Array *a) {
  int s = arraysize(a, 0);
  for (int i = 1; i < a->dim; i++)
    s *= arraysize(a, i);
  return s;
}

static int new(lua_State *L) {
  luaL_checktype(L, 1, LUA_TTABLE);
  lua_len(L, 1);
  int dim = luaL_checkint(L, -1);
  if (dim == 0)
    luaL_error(L, "invalid dimention = %d", dim);
  int size[dim];
  int bsize = sizeof(Cell);
  for (int i = 0; i < dim; i++) {
    lua_rawgeti(L, 1, i + 1);
    size[i] = luaL_checkint(L, -1);
    if (size[i] < 1)
      luaL_error(L, "invalid size %d = %d", i + 1, size[i]);
    bsize *= size[i];
  }
  bsize += sizeof(int) * (dim + 1);
  Array *a = lua_newuserdata(L, bsize);
  memset(a, 0, bsize);
  a->dim = dim;
  for (int i = 0; i < dim; i++)
    arraysize(a, i) = size[i];
  luaL_getmetatable(L, ARRAYMETA);
  lua_setmetatable(L, -2);
  return 1;
}

static int dim(lua_State *L) {
  Array *a = checkarray(L);
  lua_pushinteger(L, a->dim);
  return 1;
}

static int size(lua_State *L) {
  Array *a = checkarray(L);
  for (int i = 0; i < a->dim; i++)
    lua_pushinteger(L, arraysize(a, i));
  return a->dim;
}

Cell *cellptr(lua_State *L) {
  Array *a = checkarray(L);
  luaL_checktype(L, 2, LUA_TTABLE);
  int ix[a->dim];
  for (int i = 0; i < a->dim; i++) {
    lua_rawgeti(L, 2, i + 1);
    ix[i] = luaL_checkint(L, -1);
    if (ix[i] < 0 || ix[i] >= arraysize(a, i))
      luaL_error(L, "invalid index %d = %d", i + 1, ix[i]);
  }
  switch (a->dim) {
    case 2:
      return &arrayval(a, ix[0] * arraysize(a, 1) + ix[1]);
    default:
      luaL_error(L, "dimentions ~= 2 not yet implemented");
      return NULL; // dumb
  }
}

static int setcell(lua_State *L) {
  Cell *cell = cellptr(L);
  cell->area = luaL_checkint(L, 3);
  cell->state = luaL_checkint(L, 4);
  return 0;
}

static int setarea(lua_State *L) {
  Cell *cell = cellptr(L);
  cell->area = luaL_checkint(L, 3);
  return 0;
}

static int setstate(lua_State *L) {
  Cell *cell = cellptr(L);
  cell->state = luaL_checkint(L, 3);
  return 0;
}

static int cell(lua_State *L) {
  Cell *cell = cellptr(L);
  lua_pushinteger(L, cell->area);
  lua_pushinteger(L, cell->state);
  return 2;
}

static int area(lua_State *L) {
  Cell *cell = cellptr(L);
  lua_pushinteger(L, cell->area);
  return 1;
}

static int state(lua_State *L) {
  Cell *cell = cellptr(L);
  lua_pushinteger(L, cell->state);
  return 1;
}

static int particle(lua_State *L) {
  Cell *cell = cellptr(L);
  int bit = 1 << (luaL_checkint(L, 3) - 1);
  lua_pushinteger(L, cell->state & bit);
  return 1;
}

static int _indices(lua_State *L) {
  Array *a = _checkarray(L, lua_upvalueindex(1));
  int k = lua_tointeger(L, lua_upvalueindex(2));
  if (k >= arraylen(a)) {
    lua_pushnil(L);
    return 1;
  }
  lua_pushinteger(L, k + 1);
  lua_replace(L, lua_upvalueindex(2));
  switch (a->dim) {
    case 1:
      lua_pushinteger(L, k);
      lua_rawseti(L, lua_upvalueindex(3), 1);
      break;
    case 2:
      lua_pushinteger(L, k / arraysize(a, 1));
      lua_pushinteger(L, k % arraysize(a, 1));
      lua_rawseti(L, lua_upvalueindex(3), 2);
      lua_rawseti(L, lua_upvalueindex(3), 1);
      break;
    default:
      luaL_error(L, "dimentions > 2 not yet implemented");
  }
  lua_pushvalue(L, lua_upvalueindex(3));
  return 1;
}

static int indices(lua_State *L) {
  Array *a = checkarray(L);
  lua_pushinteger(L, 0);
  lua_createtable(L, a->dim, 0);
  for (int i = 0; i < a->dim; i++) {
    lua_pushinteger(L, 0);
    lua_rawseti(L, 3, i + 1);
  }
  lua_pushcclosure(L, &_indices, 3);
  return 1;
}

static int _values(lua_State *L) {
  Array *a = _checkarray(L, lua_upvalueindex(1));
  int k = lua_tointeger(L, lua_upvalueindex(2));
  if (k >= arraylen(a)) {
    lua_pushnil(L);
    lua_pushnil(L);
    return 2;
  }
  lua_pushinteger(L, k + 1);
  lua_replace(L, lua_upvalueindex(2));
  Cell cell = arrayval(a, k);
  lua_pushinteger(L, cell.area);
  lua_pushinteger(L, cell.state);
  return 2;
}

static int values(lua_State *L) {
  lua_pushinteger(L, 0);
  lua_pushcclosure(L, &_values, 2);
  return 1;
}

static int unpack(lua_State *L) {
  int state = luaL_checkint(L, 1);
  int count = luaL_checkint(L, 2);
  lua_createtable(L, count, 0);
  int i = 1;
  for (int k = 1; k <= count; k++) {
    int bit = state & i ? 1 : 0;
    lua_pushinteger(L, bit);
    lua_rawseti(L, -2, k);
    i <<= 1;
  }
  return 1;
}

static int propagate(lua_State *L) {
  Array *a1 = _checkarray(L, 1);
  Array *a2 = _checkarray(L, 2);
  if (a1 == a2)
    luaL_error(L, "arguments must be different");
  if (a1->dim != a2->dim)
    luaL_error(L, "unequal dimentions (%d ~= %d)", a1->dim, a2->dim);
  else
    for (int i = 0; i < a1->dim; i++)
      if (arraysize(a1, i) != arraysize(a2, i))
        luaL_error(L, "unequal size %d (%d ~= %d)", i, arraysize(a1, i), arraysize(a2, i));
  switch (a1->dim) {
    case 2: {
      int s0 = arraysize(a1, 0);
      int s1 = arraysize(a1, 1);
      Cell *c1 = arrayptr(a1, 0);
      Cell *c2 = arrayptr(a2, 0);
      for (int i0 = 0; i0 < s0; i0++) {
        for (int i1 = 0; i1 < s1; i1++) {
          int n1 = i0 + 1 == s0 ? s1 * (1 - s0) : s1;
          int n3 = i0 - 1 == -1 ? s1 * (s0 - 1) : -s1;
          int n2 = i1 - 1 == -1 ? s1 - 1 : -1;
          int n4 = i1 + 1 == s1 ? 1 - s1 : 1;
          c1->state = c2->state & ~15;
          c1->state |= c2[n1].state & 1; // down
          c1->state |= c2[n2].state & 2; // left
          c1->state |= c2[n3].state & 4; // up
          c1->state |= c2[n4].state & 8; // right
          /*
          if (c1->area != c2[n2].area && c2[n2].area != 3) {
            float r = (float) rand() / RAND_MAX;
            if (r < 0.25) {
              c1->state &= ~2;
              c1->state |= (c2->state & 8) >> 2;
            }
          }
          if (c1->area != c2[n4].area && c2[n4].area != 3) {
            float r = (float) rand() / RAND_MAX;
            if (r < 0.25) {
              c1->state &= ~8;
              c1->state |= (c2->state & 2) << 2;
            }
          }
          */
          c1++;
          c2++;
        }
      }
      break;
    }
    default:
      luaL_error(L, "dimentions ~= 2 not yet implemented");
  }
  return 1;
}

static const struct luaL_Reg libf[] = {
  {"new", new},
  {"unpack", unpack},
  {NULL, NULL},
};

static const struct luaL_Reg libm[] = {
  {"dim", dim},
  {"size", size},
  {"setcell", setcell},
  {"setarea", setarea},
  {"setstate", setstate},
  {"cell", cell},
  {"area", area},
  {"state", state},
  {"particle", particle},
  {"indices", indices},
  {"values", values},
  {"propagate", propagate},
  {NULL, NULL},
};

static void createmeta(lua_State *L) {
  luaL_newmetatable(L, ARRAYMETA);
  lua_pushvalue(L, -1);
  lua_setfield(L, -2, "__index");
  luaL_setfuncs(L, libm, 0);
  lua_pop(L, 1);
}

int luaopen_array(lua_State *L) {
  luaL_newlib(L, libf);
  createmeta(L);
  return 1;
}
