local modname = ...
local M = {}
package.loaded[modname] = M
setmetatable(M, {__index = _G})
_ENV = M

local array = require 'array'
local util = require 'util'
local view = require 'view'
local glpk = require 'glpk'

local function ipush(a, b)
    a[#a + 1] = b
end

function collide(areas)
    local rules = {}
    for i, area in ipairs(areas) do
        local ps = util.iconcat(area.mv, area.rp)
        local pc = #ps
        local maxs = bit32.extract(-1, 0, pc)

        local function masspulse(state)
            local s = array.unpack(state, pc)
            local mass = 0
            for i, p in ipairs(ps) do
                if s[i] == 1 then
                    mass = mass + p.mass
                end
            end
            local pulse = {}
            for i, _ in ipairs(area.mv[1].pulse) do
                pulse[i] = 0
            end
            for i, p in ipairs(area.mv) do
                if s[i] == 1 then
                    for j, v in ipairs(p.pulse) do
                        pulse[j] = pulse[j] + v
                    end
                end
            end
            return mass, pulse
        end

        local function equal(i, j)
            m1, p1 = masspulse(i)
            m2, p2 = masspulse(j)
            for i, v in ipairs(p2) do
                if p1[i] ~= v then
                    return false
                end
            end
            return m1 == m2
        end

        local function classrules(class)
            local a = {}
            local b = {}
            local c = {}
            local cb = {}
            local n = #class
            local m = n + (n * n - n) / 2

            for i = 1, m do
                a[i] = {}
                for j = 1, n * n do
                    a[i][j] = 0
                end
            end

            local k = 1
            for i = 1, n do
                for j = 1, n do
                    -- normalization
                    a[i][n * (i - 1) + j] = 1

                    -- gdbc
                    if i > j then
                        local _i = 0
                        local _j = 0
                        local pbit = 1
                        for l = 1, pc do
                            local b = ps[l].mass * ps[l].a
                            _i = _i + (bit32.band(class[i], pbit) == 0 and 0 or 1) * b
                            _j = _j + (bit32.band(class[j], pbit) == 0 and 0 or 1) * b
                            pbit = bit32.lshift(pbit, 1)
                        end
---[[
                        a[n + k][n * (i - 1) + j] =  math.exp(_i)
                        a[n + k][n * (j - 1) + i] = -math.exp(_j)
--]]
--[[
                        a[n + k][n * (i - 1) + j] = _i
                        a[n + k][n * (j - 1) + i] = _j
--]]
                        b[n + k] = 0
                        k = k + 1
                    end

                    -- function
                    c[n * (i - 1) + j] = i == j and 1 or 0

                    -- bounds
                    cb[n * (i - 1) + j] = {0, 1, type = 'db'}
                end
                b[i] = 1
            end

--[[
            print(unpack(class))
            for i = 1, m do
                for j = 1, n * n do
                    io.write(string.format('%10.3f', math.log(math.abs(a[i][j]))))
                end
                print('')
            end
--]]

            _, x = glpk.solve(a, c, cb, b, 'min')

            local y = {}
            for i = 1, n do
                y[i] = {}
                for j = 1, n do
                    y[i][j] = {
                        state = class[j],
                        chance = x[n * (i - 1) + j],
                    }
                end
            end
            return y
        end

        rules[i] = {}
        for j = 0, maxs do
            if not rules[i][j] then
                local class = {}
                for k = 0, maxs do
                    if equal(j, k) then
                        ipush(class, k)
                    end
                end
                local r = classrules(class)
                for k, c in ipairs(class) do
                    rules[i][c] = r[k]
                end
            end
        end
    end

    return function (area, state)
        local rule = rules[area][state]
        local chance = math.random()
        local k = 1
        while (chance > rule[k].chance) do
            chance = chance - rule[k].chance
            k = k + 1
        end
        return rule[k].state
    end, rules
end

collect = {}
local collectmeta = {
    __index = function(_, param)
        error("unknown parameter '" .. param .. "'")
    end
}
setmetatable(collect, collectmeta)

function collect.density(areas, opt)
    local pss = {}
    for i, a in ipairs(areas) do
        pss[i] = {}
        local ps = util.iconcat(a.mv, a.rp)
        for j, p in ipairs(ps) do
            pss[i][j] = p.mass
        end
    end
    local v = view.density.new(pss)

    local ints = opt.intervals
    return function(a, i)
        if a:dim() ~= 2 then
            error('invalid dimention')
        end

        for j, int in pairs(ints) do
            local f, l, s = unpack(int)
            if i > l then
                ints[j] = nil
            elseif f <= i and i <= l and (i - f) % s == 0 then
                for _, out in ipairs(opt.outputs) do
                    local filename = string.format('data/density/%s/%08d', out, i)
                    local mask = -1
                    if out == 'mv' then
                        mask = 15
                    elseif out == 'rp' then
                        mask = bit32.bxor(-1, 15)
                    end
---[[
                    filename = filename .. '.dat'
                    v:write(a, filename, mask)
--]]
--[[
                    filename = filename .. '.pgm'
                    v:writepgm(a, filename, mask)
--]]
                end
                break
            end
        end
    end
end

function collect.all(areas, params)
    local c = {}
    for param, opt in pairs(params) do
        c.param = collect[param](areas, opt)
        util.mkdir('data')
        util.mkdir('data/' .. param)
        for _, out in ipairs(opt.outputs) do
            util.mkdir('data/' .. param .. '/' .. out)
        end
    end
    return function(a, iter)
        for param, opt in pairs(params) do
            c.param(a, iter)
        end
    end
end

local function _init(lg, k, area)
    for _, g in ipairs(area.geometry) do
        for i in util.indices(table.unpack(g)) do
            local s = util.iconcat(area.mv, area.rp)
            for i, p in ipairs(s) do
                s[i] = math.random() < p.appear and 1 or 0
            end
            local ps = array.pack(s)
            lg[1]:setcell(i, k, ps)
            lg[2]:setcell(i, k, ps)
        end
    end
end

function new(model)
    local lg = {
        array.new(model.size),
        array.new(model.size),
    }
    local rules = nil
    lg.collide, rules = collide(model.areas)
    setmetatable(lg, {__index = M})
    for i, area in ipairs(model.areas) do
        _init(lg, i, area)
    end
    lg.array = lg[1]
    return lg, rules
end

function step(self)
    self[2]:propagate(self[1])
    for i in self[1]:indices() do
        self[1]:setstate(i, self.collide(self[2]:cell(i)))
    end
end
