#define _XOPEN_SOURCE 700

#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <lauxlib.h>
#include <ftw.h>

static int iclone(lua_State *L) {
    luaL_checktype(L, 1, LUA_TTABLE);
    lua_len(L, 1);
    int size = luaL_checkint(L, -1);
    lua_createtable(L, size, 0);
    for (int i = 1; i <= size; i++) {
        lua_rawgeti(L, 1, i);
        lua_rawseti(L, -2, i);
    }
    return 1;
}

static int iconcat(lua_State *L) {
    luaL_checktype(L, 1, LUA_TTABLE);
    lua_len(L, 1);
    int size_1 = luaL_checkint(L, -1);
    luaL_checktype(L, 2, LUA_TTABLE);
    lua_len(L, 2);
    int size = size_1 + luaL_checkint(L, -1);
    lua_createtable(L, size, 0);
    for (int i = 1; i <= size_1; i++) {
        lua_rawgeti(L, 1, i);
        lua_rawseti(L, -2, i);
    }
    for (int i = size_1 + 1; i <= size; i++) {
        lua_rawgeti(L, 2, i - size_1);
        lua_rawseti(L, -2, i);
    }
    return 1;
}

static int makedir(lua_State *L) {
    const char *path = luaL_checkstring(L, 1);
    mode_t mode = S_IRUSR | S_IWUSR | S_IXUSR | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH;
    if (mkdir(path, mode)) {
        lua_pushnil(L);
        lua_pushfstring(L, "%s", strerror(errno));
        return 2;
    }
    lua_pushboolean(L, 1);
    return 1;
}

static int changedir(lua_State *L) {
    const char *path = luaL_checkstring(L, 1);
    if (chdir(path)) {
        lua_pushnil(L);
        lua_pushfstring(L, "%s", strerror(errno));
        return 2;
    }
    lua_pushboolean(L, 1);
    return 1;
}

static int level;

int unlinkcb(const char *path, const struct stat *sb, int typeflag, struct FTW *ftwbuf) {
    if (ftwbuf->level <= level)
        return 0;
    return remove(path);
}

static int rm(lua_State *L) {
    const char *path = luaL_checkstring(L, 1);
    int status = 1;
    level = -1;
    if (lua_gettop(L) > 1) {
        const char *mod = luaL_checkstring(L, 2);
        if (lua_gettop(L) > 2)
            level = luaL_checkint(L, 3);
        if (!strncmp(mod, "r", 1))
            status = nftw(path, unlinkcb, 64, FTW_DEPTH | FTW_PHYS);
        else
            luaL_error(L, "invalid modifier");
    }
    else
        status = remove(path);
    if (status) {
        lua_pushnil(L);
        lua_pushfstring(L, "%s", strerror(errno));
        return 2;
    }
    lua_pushboolean(L, 1);
    return 1;
}

static int getdir(lua_State *L) {
    char *path;
    char buf[PATH_MAX];
    path = getcwd(buf, PATH_MAX);
    if (path) {
        lua_pushstring(L, path);
        return 1;
    }
    else {
        lua_pushnil(L);
        lua_pushfstring(L, "%s", strerror(errno));
        return 2;
    }
}

static const struct luaL_Reg libf[] = {
    {"iclone", iclone},
    {"iconcat", iconcat},
    {"mkdir", makedir},
    {"chdir", changedir},
    {"rm", rm},
    {"dir", getdir},
    {NULL, NULL},
};

int luaopen_util(lua_State *L) {
    luaL_newlib(L, libf);
    return 1;
}
