#include "array.h"

#include <string.h>
#include <errno.h>
#include <lauxlib.h>
#include <stdlib.h>

const int K = 5;
const int Y = 11;
const int C = 3;

//#define FAVG

typedef struct {
    int amount;
    int maxsize;
    char _[0];
} Density;

#define max(a, b) ((a) > (b) ? a : b)

#define DENSITYMETA "densitymeta"

#define densitysize(d, i) *((int *) (d->_ + sizeof(int) * (i)))
#define densityarea(d, i) (d->_ + sizeof(int) * d->amount + d->maxsize * (i))

#define _checkdensity(L, i) luaL_checkudata(L, i, DENSITYMETA)
#define checkdensity(L) _checkdensity(L, 1)

static int densitynew(lua_State *L) {
    luaL_checktype(L, 1, LUA_TTABLE);
    lua_len(L, 1);
    int amount = luaL_checkint(L, -1);
    lua_pop(L, 1);
    if (amount == 0)
        luaL_error(L, "invalid amount of areas = %d", amount);
    int _size[amount];
    int size[amount];
    int maxsize = 0;
    for (int i = 0; i < amount; i++) {
        lua_rawgeti(L, 1, i + 1);
        lua_len(L, -1);
        _size[i] = lua_tointeger(L, -1);
        size[i] = 1 << _size[i];
        lua_pop(L, 1);
        if (_size[i] == 0)
            luaL_error(L, "invalid amount of particles %d = %d", i + 1, size[i]);
        if (maxsize < size[i])
            maxsize = size[i];
    }
    int bsize = maxsize * amount + sizeof(int) * (amount + 2);
    Density *d = lua_newuserdata(L, bsize);
    memset(d, 0, bsize);
    d->amount = amount;
    d->maxsize = maxsize;
    for (int i = 0; i < amount; i++) {
        densitysize(d, i) = size[i];
        lua_rawgeti(L, 1, i + 1);
        int table = lua_gettop(L);
        for (int j = 0; j < size[i]; j++) {
            int bit = 1;
            int dval = 0;
            for (int k = 0; k < _size[i]; k++) {
                if (j & bit) {
                    lua_rawgeti(L, table, k + 1);
                    dval += lua_tointeger(L, -1);
                    lua_pop(L, 1);
                }
                bit <<= 1;
            }
            densityarea(d, i)[j] = dval;
        }
        lua_pop(L, 1);
    }
    luaL_getmetatable(L, DENSITYMETA);
    lua_setmetatable(L, -2);
    return 1;
}

#if 1

static void smooth(float *a, int s) {
    float *b = calloc(s, sizeof(float));
    for (int i = -K; i <= K; i++)
        b[0] += a[(i + s) % s];
    for (int i = 1; i < s; i++)
        b[i] = b[i - 1] - a[(i - 1 - K + s) % s] + a[(i + K) % s];
    for (int i = 0; i < s; i++)
        a[i] = b[i] / (K + K + 1);
    free(b);
}

static int densitywrite(lua_State *L) {
    Density *d = _checkdensity(L, 1);
    Array *a = _checkarray(L, 2);
    const char *path = luaL_checkstring(L, 3);
    State mask = luaL_checkint(L, 4);
    int vertical = lua_toboolean(L, 5);
    FILE *fp = fopen(path, "w");
    if (!fp) {
        lua_pushnil(L);
        lua_pushfstring(L, "%s", strerror(errno));
        return 2;
    }
    switch (a->dim) {
        case 2: {
                int s0 = arraysize(a, 0);
                int s1 = arraysize(a, 1);
                int s = vertical ? s0 : s1;
                float *val = calloc(s, sizeof(float));
#ifdef FAVG
                for (int i0 = 0; i0 < s0; i0++) {
                    for (int i1 = 0; i1 < s1; i1++) {
                        Cell c = arrayval(a, i0 * s1 + i1);
                        int i = vertical ? i0 : i1;
                        val[i] += densityarea(d, c.area - 1)[c.state & mask];
                    }
                }
#else
                for (int k = -Y; k <= Y; k++) {
                    for (int i = 0; i < s; i++) {
                        int i0 = vertical ? i : s0 / 2 + k;
                        int i1 = vertical ? s1 / 2 + k : i;
                        Cell c = arrayval(a, i0 * s1 + i1);
                        val[i] += densityarea(d, c.area - 1)[c.state & mask];
                    }
                }
#endif
                const int count = C;
                for (int i = 0; i < count; i++)
                    smooth(val, s);

                for (int i = 0; i < s; i++)
#ifdef FAVG
                    fprintf(fp, "%f\n", val[i] / (s0 + s1 - s));
#else
                    fprintf(fp, "%f\n", val[i] / (Y + Y + 1));
#endif
                free(val);
        } break;
        default:
            fclose(fp);
            luaL_error(L, "dimentions ~= 2 not yet implemented");
    }
    fclose(fp);
    lua_pushboolean(L, 1);
    return 1;
}

#else

static void convolution_(float *a, int n, int m) {
    const float r = 3;
    for (int i = 0; i < n; i++)
        for (int j = 0; j < m; j++) {
            float s = 0;
            for (int k = -r; k <= r; k++) {
                int t = (j + k + m) % m;
                s += a[m * i + t];
            }
            a[m * i + j] = s / (r + r + 1);
        }
    for (int j = 0; j < m; j++)
        for (int i = 0; i < n; i++) {
            float s = 0;
            for (int k = -r; k <= r; k++) {
                int t = (i + k + n) % n;
                s += a[m * t + j];
            }
            a[m * i + j] = s / (r + r + 1);
        }
}

static int densitywrite(lua_State *L) {
    Density *d = _checkdensity(L, 1);
    Array *a = _checkarray(L, 2);
    const char *path = luaL_checkstring(L, 3);
    State mask = luaL_checkint(L, 4);
    FILE *fp = fopen(path, "w");
    if (!fp) {
        lua_pushnil(L);
        lua_pushfstring(L, "%s", strerror(errno));
        return 2;
    }
    switch (a->dim) {
        case 2: {
            int s0 = arraysize(a, 0);
            int s1 = arraysize(a, 1);
            Cell *c = arrayptr(a, 0);
//            int k = 1;
            float val[s0 * s1];
            for (int i0 = 0; i0 < s0; i0++)
                for (int i1 = 0; i1 < s1; i1++) {
                    val[s1 * i0 + i1] = densityarea(d, c->area - 1)[c->state & mask];
//                    k = max(k, val[s1 * i0 + i1]);
                    c++;
                }
            /*
            k = 255 / k;
            for (int i0 = 0; i0 < s0; i0++)
                for (int i1 = 0; i1 < s1; i1++)
                    val[s1 * i0 + i1] *= k;
            */
            const int count = 4;
            for (int i = 0; i < count; i++)
                convolution_(val, s0, s1);
//            fprintf(fp, "P5\n%d\n%d\n%d\n", s1, s0, 255);
//            fwrite(val + s0 / 2 * s1, 1, s1, fp);
            for (int i = 0; i < s1; i++)
                fprintf(fp, "%f\n", val[(s0 / 2) * s1 + i]);
        } break;
        default:
            fclose(fp);
            luaL_error(L, "dimentions ~= 2 not yet implemented");
    }
    fclose(fp);
    lua_pushboolean(L, 1);
    return 1;
}
#endif

static void convolution(unsigned char *a, int n, int m) {
    const int r = 1;
    for (int i = 0; i < n; i++)
        for (int j = 0; j < m; j++) {
            int s = 0;
            for (int k = -r; k <= r; k++) {
                int t = (j + k + m) % m;
                s += a[m * i + t];
            }
            a[m * i + j] = s / (r + r + 1);
        }
    for (int j = 0; j < m; j++)
        for (int i = 0; i < n; i++) {
            int s = 0;
            for (int k = -r; k <= r; k++) {
                int t = (i + k + n) % n;
                s += a[m * t + j];
            }
            a[m * i + j] = s / (r + r + 1);
        }
}

static int densitywritepgm(lua_State *L) {
    Density *d = _checkdensity(L, 1);
    Array *a = _checkarray(L, 2);
    const char *path = luaL_checkstring(L, 3);
    State mask = luaL_checkint(L, 4);
    FILE *fp = fopen(path, "w");
    if (!fp) {
        lua_pushnil(L);
        lua_pushfstring(L, "%s", strerror(errno));
        return 2;
    }
    switch (a->dim) {
        case 2: {
            int s0 = arraysize(a, 0);
            int s1 = arraysize(a, 1);
            Cell *c = arrayptr(a, 0);
            int k = 1;
            unsigned char val[s0 * s1];
            for (int i0 = 0; i0 < s0; i0++)
                for (int i1 = 0; i1 < s1; i1++) {
                    val[s1 * i0 + i1] = densityarea(d, c->area - 1)[c->state & mask];
                    k = max(k, val[s1 * i0 + i1]);
                    c++;
                }
            k = 255 / k;
            for (int i0 = 0; i0 < s0; i0++)
                for (int i1 = 0; i1 < s1; i1++)
                    val[s1 * i0 + i1] *= k;
            const int count = 1;
            for (int i = 0; i < count; i++)
                convolution(val, s0, s1);
            fprintf(fp, "P5\n%d\n%d\n%d\n", s1, s0, 255);
            fwrite(val, 1, s0 * s1, fp);
        } break;
        default:
            fclose(fp);
            luaL_error(L, "dimentions ~= 2 not yet implemented");
    }
    fclose(fp);
    lua_pushboolean(L, 1);
    return 1;
}

static const struct luaL_Reg densityf[] = {

    {"new", densitynew},
    {NULL, NULL},
};

static const struct luaL_Reg densitym[] = {
    {"write", densitywrite},
    {"writepgm", densitywritepgm},
    {NULL, NULL},
};

static void createdensity(lua_State *L) {
    luaL_newlib(L, densityf);
    luaL_newmetatable(L, DENSITYMETA);
    lua_pushvalue(L, -1);
    lua_setfield(L, -2, "__index");
    luaL_setfuncs(L, densitym, 0);
    lua_pop(L, 1);
}

int luaopen_view(lua_State *L) {
    lua_createtable(L, 0, 1);
    createdensity(L);
    lua_setfield(L, -2, "density");
    return 1;
}
